# coding: utf-8
#
#	jma_icon.rb: JMA weather icon plugin for tDiary (UTF-8)
#
#	Copyright (C) 2021 by Gaju Westwood <gaju@gaju.jp>
#	Distributed under GPL.
#
#  2021.02.26 initial revision
#  2021.02.27 fix conversion of weather code to icon file
#  2021.03.02 added semi-auto fetch feature
#  2021.03.04 use title attribute to write summary
#

require 'open-uri'
require 'timeout'
require 'json'
require 'date'
require 'time'
require 'fileutils'

#
#	Constants
#
JmaIcon_default_url     = 'https://www.jma.go.jp/bosai/forecast/data/forecast'
JmaIcon_default_img_dir = 'https://www.jma.go.jp/bosai/forecast/img'
JmaIcon_default_station = '130000'
JmaIcon_default_area    = '130010'
JmaIcon_default_dir     = "#{@cache_path}/jma_icon"
JmaIcon_code_to_icon = {
	"100" => [ "100.svg", "晴" ],
	"101" => [ "101.svg", "晴時々曇" ],
	"102" => [ "102.svg", "晴一時雨" ],
	"103" => [ "102.svg", "晴時々雨" ],
	"104" => [ "104.svg", "晴一時雪" ],
	"105" => [ "104.svg", "晴時々雪" ],
	"106" => [ "102.svg", "晴一時雨か雪" ],
	"107" => [ "102.svg", "晴時々雨か雪" ],
	"108" => [ "102.svg", "晴一時雨か雷雨" ],
	"110" => [ "110.svg", "晴後時々曇" ],
	"111" => [ "110.svg", "晴後曇" ],
	"112" => [ "112.svg", "晴後一時雨" ],
	"113" => [ "112.svg", "晴後時々雨" ],
	"114" => [ "112.svg", "晴後雨" ],
	"115" => [ "115.svg", "晴後一時雪" ],
	"116" => [ "115.svg", "晴後時々雪" ],
	"117" => [ "115.svg", "晴後雪" ],
	"118" => [ "112.svg", "晴後雨か雪" ],
	"119" => [ "112.svg", "晴後雨か雷雨" ],
	"120" => [ "102.svg", "晴朝夕一時雨" ],
	"121" => [ "102.svg", "晴朝の内一時雨" ],
	"122" => [ "112.svg", "晴夕方一時雨" ],
	"123" => [ "100.svg", "晴山沿い雷雨" ],
	"124" => [ "100.svg", "晴山沿い雪" ],
	"125" => [ "112.svg", "晴午後は雷雨" ],
	"126" => [ "112.svg", "晴昼頃から雨" ],
	"127" => [ "112.svg", "晴夕方から雨" ],
	"128" => [ "112.svg", "晴夜は雨" ],
	"130" => [ "100.svg", "朝の内霧後晴" ],
	"131" => [ "100.svg", "晴明け方霧" ],
	"132" => [ "101.svg", "晴朝夕曇" ],
	"140" => [ "102.svg", "晴時々雨で雷を伴う" ],
	"160" => [ "104.svg", "晴一時雪か雨" ],
	"170" => [ "104.svg", "晴時々雪か雨" ],
	"181" => [ "115.svg", "晴後雪か雨" ],
	"200" => [ "200.svg", "曇" ],
	"201" => [ "201.svg", "曇時々晴" ],
	"202" => [ "202.svg", "曇一時雨" ],
	"203" => [ "202.svg", "曇時々雨" ],
	"204" => [ "204.svg", "曇一時雪" ],
	"205" => [ "204.svg", "曇時々雪" ],
	"206" => [ "202.svg", "曇一時雨か雪" ],
	"207" => [ "202.svg", "曇時々雨か雪" ],
	"208" => [ "202.svg", "曇一時雨か雷雨" ],
	"209" => [ "200.svg", "霧" ],
	"210" => [ "210.svg", "曇後時々晴" ],
	"211" => [ "210.svg", "曇後晴" ],
	"212" => [ "212.svg", "曇後一時雨" ],
	"213" => [ "212.svg", "曇後時々雨" ],
	"214" => [ "212.svg", "曇後雨" ],
	"215" => [ "215.svg", "曇後一時雪" ],
	"216" => [ "215.svg", "曇後時々雪" ],
	"217" => [ "215.svg", "曇後雪" ],
	"218" => [ "212.svg", "曇後雨か雪" ],
	"219" => [ "212.svg", "曇後雨か雷雨" ],
	"220" => [ "202.svg", "曇朝夕一時雨" ],
	"221" => [ "202.svg", "曇朝の内一時雨" ],
	"222" => [ "212.svg", "曇夕方一時雨" ],
	"223" => [ "201.svg", "曇日中時々晴" ],
	"224" => [ "212.svg", "曇昼頃から雨" ],
	"225" => [ "212.svg", "曇夕方から雨" ],
	"226" => [ "212.svg", "曇夜は雨" ],
	"228" => [ "215.svg", "曇昼頃から雪" ],
	"229" => [ "215.svg", "曇夕方から雪" ],
	"230" => [ "215.svg", "曇夜は雪" ],
	"231" => [ "200.svg", "曇海上海岸は霧か霧雨" ],
	"240" => [ "202.svg", "曇時々雨で雷を伴う" ],
	"250" => [ "204.svg", "曇時々雪で雷を伴う" ],
	"260" => [ "204.svg", "曇一時雪か雨" ],
	"270" => [ "204.svg", "曇時々雪か雨" ],
	"281" => [ "215.svg", "曇後雪か雨" ],
	"300" => [ "300.svg", "雨" ],
	"301" => [ "301.svg", "雨時々晴" ],
	"302" => [ "302.svg", "雨時々止む" ],
	"303" => [ "303.svg", "雨時々雪" ],
	"304" => [ "300.svg", "雨か雪" ],
	"306" => [ "300.svg", "大雨" ],
	"308" => [ "308.svg", "雨で暴風を伴う" ],
	"309" => [ "303.svg", "雨一時雪" ],
	"311" => [ "311.svg", "雨後晴" ],
	"313" => [ "313.svg", "雨後曇" ],
	"314" => [ "314.svg", "雨後時々雪" ],
	"315" => [ "314.svg", "雨後雪" ],
	"316" => [ "311.svg", "雨か雪後晴" ],
	"317" => [ "313.svg", "雨か雪後曇" ],
	"320" => [ "311.svg", "朝の内雨後晴" ],
	"321" => [ "313.svg", "朝の内雨後曇" ],
	"322" => [ "303.svg", "雨朝晩一時雪" ],
	"323" => [ "311.svg", "雨昼頃から晴" ],
	"324" => [ "311.svg", "雨夕方から晴" ],
	"325" => [ "311.svg", "雨夜は晴" ],
	"326" => [ "314.svg", "雨夕方から雪" ],
	"327" => [ "314.svg", "雨夜は雪" ],
	"328" => [ "300.svg", "雨一時強く降る" ],
	"329" => [ "300.svg", "雨一時みぞれ" ],
	"340" => [ "400.svg", "雪か雨" ],
	"350" => [ "300.svg", "雨で雷を伴う" ],
	"361" => [ "411.svg", "雪か雨後晴" ],
	"371" => [ "413.svg", "雪か雨後曇" ],
	"400" => [ "400.svg", "雪" ],
	"401" => [ "401.svg", "雪時々晴" ],
	"402" => [ "402.svg", "雪時々止む" ],
	"403" => [ "403.svg", "雪時々雨" ],
	"405" => [ "400.svg", "大雪" ],
	"406" => [ "406.svg", "風雪強い" ],
	"407" => [ "406.svg", "暴風雪" ],
	"409" => [ "403.svg", "雪一時雨" ],
	"411" => [ "411.svg", "雪後晴" ],
	"413" => [ "413.svg", "雪後曇" ],
	"414" => [ "414.svg", "雪後雨" ],
	"420" => [ "411.svg", "朝の内雪後晴" ],
	"421" => [ "413.svg", "朝の内雪後曇" ],
	"422" => [ "414.svg", "雪昼頃から雨" ],
	"423" => [ "414.svg", "雪夕方から雨" ],
	"425" => [ "400.svg", "雪一時強く降る" ],
	"426" => [ "400.svg", "雪後みぞれ" ],
	"427" => [ "400.svg", "雪一時みぞれ" ],
	"450" => [ "400.svg", "雪で雷を伴う" ]
}

#
#	CLASS: JmaIcon
#
class JmaIcon
	attr_reader :date, :error, :unmatch, :data

	HTML_START = '<div class="jma">'
	HTML_END   = '</div>'

	WAITTIME = 10

	def get_date_index( req_date, json_doc, series = 0, step = 1 )
		idx = -1
		json_doc[0]['timeSeries'][series]['timeDefines'].each do |date_str|
			idx += 1
			if ( idx % step ) != 0 then
				next
			end
			date = Time.parse( date_str )
			if date.hour == 0 then
				get_date = JmaIcon.date_to_s( date )
				if req_date == get_date then
					return idx
				end
			end
		end
		return -1
	end

	def get_area_index( req_area, json_doc )
		idx = -1
		json_doc[0]['timeSeries'][0]['areas'].each do |area_data|
			idx += 1
			if req_area == area_data['area']['code'] then
				return idx
			end
		end
		return -1
	end

	def parse_json( conf, json_doc )
		@unmatch = true
		return if not json_doc

		req_date = JmaIcon.date_to_s( @date )
		date_idx = get_date_index( req_date, json_doc )
		return if date_idx < 0

		area_idx = get_area_index( conf['jma_icon.area'], json_doc )
		return if area_idx < 0

		@unmatch = nil
		@data['time'] = conf.to_native( json_doc[0]['timeSeries'][0]['timeDefines'][date_idx] ) rescue nil
		@data['code'] = conf.to_native( json_doc[0]['timeSeries'][0]['areas'][area_idx]['weatherCodes'][date_idx] ) rescue nil
		@data['city'] = conf.to_native( json_doc[0]['timeSeries'][2]['areas'][area_idx]['area']['name'] ) rescue nil
		@data['weather'] = conf.to_native( json_doc[0]['timeSeries'][0]['areas'][area_idx]['weathers'][date_idx] ) rescue nil
		temp_date_idx = get_date_index( req_date, json_doc, 2, 2 )
		if temp_date_idx >= 0 then
			@data['t_min'] = conf.to_native( json_doc[0]['timeSeries'][2]['areas'][area_idx]['temps'][temp_date_idx] ) rescue nil
			@data['t_max'] = conf.to_native( json_doc[0]['timeSeries'][2]['areas'][area_idx]['temps'][temp_date_idx + 1] ) rescue nil
		end
	end

	def initialize( date = nil )
		@date = date or Time.now
		@data = Hash.new
		@error = nil
		@unmatch = nil
	end

	def get( conf )
		station = conf['jma_icon.station']
		url =  conf['jma_icon.url'].gsub(/[\t\n]/, '')
		url =~ %r!https://([^/]+)(.*)!
		url =  %Q!https://#{$1}#{$2}/#{station}.json!
		@error = nil
		@unmatch = nil
		json_file = nil
		begin
			timeout( WAITTIME ) do
				proxy = conf['proxy']? ( 'http://' + conf['proxy'] ) : nil
				json_file = open( url, :proxy => proxy ) {|f| f.read }
			end
			json_doc = JSON.parse( json_file )
			parse_json( conf, json_doc )
		rescue Timeout::Error, JSON::ParserError, StandardError
			@error = conf.to_native( $!.message.gsub( /[\t\n]/, ' ' ) )
			@unmatch = true
		end
		if @unmatch then
			begin
				json_cache_file = open_cache( conf['jma_icon.dir'] )
				json_doc = JSON.parse( json_cache_file )
				parse_json( conf, json_doc )
			rescue
				# ignore cache exceptions
			end
		end
		if not @error and json_file then
			save_cache( conf['jma_icon.dir'], json_file )
		end
		self
	end

	def to_s
		r = "#{JmaIcon.date_to_s( @date )}\t#{@error}"
		@data.each do | item, value |
			r << "\t#{item}\t#{value}" if not (value||'').empty?
		end
		r << "\n"
	end

	def parse( data_line )
		i = data_line.chomp.split( /\t/ )
		y, m, d = i.shift.scan( /^(\d{4})(\d\d)(\d\d)$/ )[0]
		@date = Time.local( y, m, d )
		error = i.shift
		if not (error||'').empty? then
			@error = error
		else
			@error = nil
		end
		@data.clear
		while not i.empty? do
			@data[i.shift] = i.shift
		end
		self
	end

	def to_html( conf )
		@error ? ( conf['show_error'] ? error_html_string : '' ) : html_string( conf )
	end

	def store( dir, date )
		if not @unmatch
			data_dir = File.dirname( JmaIcon.file_path( dir, date ) )
			FileUtils.mkdir_p( data_dir )
			File.open( JmaIcon.file_path( dir, date ), 'a' ) do |fh|
				fh.puts( to_s )
			end
		end
	end

	def open_cache( dir )
		File.open( JmaIcon.cache_file_path( dir ) ) do |fh|
			fh.read
		end
	end

	def save_cache( dir, json_file )
		data_dir = File.dirname( JmaIcon.cache_file_path( dir ) )
		FileUtils.mkdir_p( data_dir )
		File.open( JmaIcon.cache_file_path( dir ), 'w' ) do |fh|
			fh.write( json_file )
		end
	end

	class << self
		def parse( data_line )
			new.parse( data_line )
		end

		def date_to_s( date )
			date.strftime( '%Y%m%d' )
		end

		def file_path( dir, date )
			date.strftime( "#{dir}/%Y/%Y%m.jma" ).gsub( /\/\/+/, '/' )
		end

		def cache_file_path( dir )
			"#{dir}/cache.json".gsub( /\/\/+/, '/' )
		end

		def agt_file_path( dir )
			"#{dir}/auto_get_time".gsub( /\/\/+/, '/' )
		end

		def restore( dir, date )
			r = nil
			req_date = JmaIcon.date_to_s( date )
			begin
				File.open( file_path( dir, date ), 'r' ) do |fh|
					fh.each( "\n" ) do |l|
						if /^#{req_date}\t/ =~ l then
							r = l
						end
					end
				end
			rescue Errno::ENOENT
			end
			r ? JmaIcon.parse( r ) : nil
		end
	end
end

#
#	CLASS: JmaIcon (display methods)
#
class JmaIcon
	include ERB::Util

	def error_html_string
		%Q!#{HTML_START}Error: #{h( @error )}#{HTML_END}!
	end

	def html_string( conf )
		show_icon = conf['jma_icon.view_type'] == 'icon' and not conf['jma_icon.img_dir'].empty? and not (@data['code']||'').empty? 
		telop = JmaIcon_code_to_icon[@data['code']][1] rescue ''
		summary = @data['weather'].gsub( /　/, ' ' ) if @data['weather']
		if conf['jma_icon.view_type'] == 'telop' and not telop.empty? then
			summary = telop
		end
		if show_icon or conf['jma_icon.show_city'] then
			city = @data['city']
			summary = city + ' - ' + summary if not city.empty? and summary
		end
		if show_icon or conf['jma_icon.show_temp'] then
			t_max = @data['t_max'] ? ( @data['t_max'] + conf['jma_icon.temp_unit'] ) : '-'
			t_min = @data['t_min'] ? ( @data['t_min'] + conf['jma_icon.temp_unit'] ) : '-'
			temp = '(' + t_max + '/' + t_min + ')'
			summary = summary + ' ' + temp if summary
		end
		r = "#{HTML_START}"
		if show_icon then
			begin
				r << city << ' ' if conf['jma_icon.show_city'] and not (city||'').empty?
				img_w = conf['jma_icon.img_width'] ? conf['jma_icon.img_width'].to_i : nil
				img_h = conf['jma_icon.img_height'] ? conf['jma_icon.img_height'].to_i : nil
				img_u = conf['jma_icon.img_dir'] + '/' + JmaIcon_code_to_icon[@data['code']][0]
				r << %Q!<img src="#{h( img_u )}"!
				r << %Q! width="#{h( img_w )}" height="#{h( img_h )}"! if img_w and img_h
				r << %Q! alt="#{h( telop )}"! if telop
				r << %Q! title="#{h( summary )}"! if summary
				r << '>'
				r << ' ' << temp if conf['jma_icon.show_temp'] and not (temp||'').empty?
			rescue
				if summary then
					r << h( summary )
				end
			end
		elsif summary then
			r << h( summary )
		end
		r << "#{HTML_END}\n"
	end
end

#
#	functions
#
def jma_icon( date = nil )
	return '' if defined? bot? and bot?
	return '' if defined? feed? and feed?
	jma_icon_init
	w = JmaIcon.restore( @conf['jma_icon.dir'], date || @date )
	if w then
		w.to_html( @conf )
	else
		''
	end
end

def jma_icon_get
	return unless @mode == 'append' or @mode == 'replace'
	jma_icon_init
	date_diff = Date.today - Date.new( @date.year, @date.mon, @date.mday )
	return unless date_diff == 0 or date_diff == 1
	w = JmaIcon.restore( @conf['jma_icon.dir'], @date )
	if not w or w.error then
		w = JmaIcon.new( @date )
		w.get( @conf )
		w.store( @conf['jma_icon.dir'], @date )
	end
end

def jma_icon_auto_get
	return unless @conf['jma_icon.auto_get']
	return unless @mode == 'latest'
	time_now = Time.now.getlocal("+09:00")
	return unless time_now.hour > 17

	jma_icon_init
	agt_file_path = JmaIcon.agt_file_path( @conf['jma_icon.dir'] )
	begin
		time_agt = File.mtime( agt_file_path ).getlocal("+09:00")
	rescue
		time_agt = nil
	end
	if time_agt then
		date_diff = Date.today - Date.new( time_agt.year, time_agt.mon, time_agt.mday )
		return unless date_diff > 0
	end

	begin
		FileUtils.touch( agt_file_path )
	rescue
		return
	end

	w = JmaIcon.restore( @conf['jma_icon.dir'], time_now )
	if not w or w.error then
		w = JmaIcon.new( time_now )
		w.get( @conf )
		w.store( @conf['jma_icon.dir'], time_now )
	end
end

def jma_icon_init
	@conf['jma_icon.url'] ||= JmaIcon_default_url
	@conf['jma_icon.station'] ||= JmaIcon_default_station
	@conf['jma_icon.area'] ||= JmaIcon_default_area
	@conf['jma_icon.img_width'] ||= '52'
	@conf['jma_icon.img_height'] ||= '39'
	@conf['jma_icon.img_dir'] ||= JmaIcon_default_img_dir
	@conf['jma_icon.view_type'] ||= 'icon'
	@conf['jma_icon.show_city'] = true if @conf['jma_icon.show_city'].nil?
	@conf['jma_icon.show_temp'] = false if @conf['jma_icon.show_temp'].nil?
	@conf['jma_icon.auto_get'] = false if @conf['jma_icon.auto_get'].nil?
	@conf['jma_icon.show_error'] = false if @conf['jma_icon.show_error'].nil?

	# maybe use default
	@conf['jma_icon.dir'] ||= JmaIcon_default_dir
	# always use default
	@conf['jma_icon.temp_unit'] = @jma_icon_temp_unit
end

def saveconf_jma_icon
	if @mode == 'saveconf' then
		@conf['jma_icon.url'] = @cgi.params['jma_icon.url'][0]
		@conf['jma_icon.station'] = @cgi.params['jma_icon.station'][0]
		@conf['jma_icon.area'] = @cgi.params['jma_icon.area'][0]
		@conf['jma_icon.img_width'] = @cgi.params['jma_icon.img_width'][0]
		@conf['jma_icon.img_height'] = @cgi.params['jma_icon.img_height'][0]
		@conf['jma_icon.img_dir'] = @cgi.params['jma_icon.img_dir'][0]
		@conf['jma_icon.view_type'] = @cgi.params['jma_icon.view_type'][0]
		@conf['jma_icon.show_city'] = @cgi.params['jma_icon.show_city'][0] ? true : false
		@conf['jma_icon.show_temp'] = @cgi.params['jma_icon.show_temp'][0] ? true : false
		@conf['jma_icon.auto_get'] = @cgi.params['jma_icon.auto_get'][0] ? true : false
		@conf['jma_icon.show_error'] = @cgi.params['jma_icon.show_error'][0] ? true : false
	end
end

def configure_jma_icon
	saveconf_jma_icon
	jma_icon_init
	jma_icon_configure_html
end

def jma_icon_configure_html
	<<-HTML
	<h3 class="subtitle">#{@jma_icon_plugin_name}</h3>
	<p>#{@jma_icon_plugin_description}</p>
	<h3 class="subtitle">#{@jma_icon_url_label}</h3>
	<p><input type="text" name="jma_icon.url" value="#{h @conf['jma_icon.url']}" size="80"></p>
	<p><input type="checkbox" id="jma_icon.auto_get" name="jma_icon.auto_get" value="true"#{" checked" if @conf['jma_icon.auto_get']}><label for="jma_icon.auto_get">#{@jma_icon_auto_get_label}</label></p>
	<h3 class="subtitle">#{@jma_icon_station_label}</h3>
	<p><input type="text" name="jma_icon.station" value="#{h @conf['jma_icon.station']}" size="5"></p>
	<h3 class="subtitle">#{@jma_icon_area_label}</h3>
	<p><input type="text" name="jma_icon.area" value="#{h @conf['jma_icon.area']}" size="5"></p>
	<h3 class="subtitle">#{@jma_icon_disp_conf_label}</h3>
	<p>#{@jma_icon_view_type_label}
		<select name="jma_icon.view_type">
		<option value="icon"#{' selected' if @conf['jma_icon.view_type'] == 'icon'}>#{@jma_icon_icon_label}</option>
		<option value="summary"#{' selected' if @conf['jma_icon.view_type'] == 'summary'}>#{@jma_icon_summary_label}</option>
		<option value="telop"#{' selected' if @conf['jma_icon.view_type'] == 'telop'}>#{@jma_icon_telop_label}</option>
		</select>
	</p>
   <p>#{@jma_icon_img_size_label} <input type="text" name="jma_icon.img_width" value="#{h @conf['jma_icon.img_width']}" size="5"> x <input type="text" name="jma_icon.img_height" value="#{h @conf['jma_icon.img_height']}" size="5"></p>
	<p>#{@jma_icon_img_dir_label} <input type="text" name="jma_icon.img_dir" value="#{h @conf['jma_icon.img_dir']}" size="80"></p>
	<p><input type="checkbox" id="jma_icon.show_city" name="jma_icon.show_city" value="true"#{" checked" if @conf['jma_icon.show_city']}><label for="jma_icon.show_city">#{@jma_icon_show_city_label}</label></p>
	<p><input type="checkbox" id="jma_icon.show_temp" name="jma_icon.show_temp" value="true"#{" checked" if @conf['jma_icon.show_temp']}><label for="jma_icon.show_temp">#{@jma_icon_show_temp_label}</label></p>
	HTML
end

#
#	hooks
#
add_body_enter_proc do |date| jma_icon( date ) end
add_update_proc do jma_icon_get end
add_conf_proc( 'jma_icon', @jma_icon_plugin_name ) do configure_jma_icon end
add_header_proc do jma_icon_auto_get end
